/*-
 * Copyright (c) 2003, Patrick MARIE <mycroft@virgaria.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD$
 */
/*
  Syntax of configuration file: /usr/local/etc/sew.conf

login:loginshell:after:there:is:commands:allowed

ex:
user1:/bin/sh:ls:date:ps
user2:/bin/sh:all

allaction is a magic keyword to allow to run a shell (normal ssh execution)
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>

#include <sys/types.h>
#include <pwd.h>
#include <syslog.h>

#define SEV_FILE "/usr/local/etc/sew.conf"

#define DEF_DENY_MSG "Access denied for user.\n"

#define ALL				"all"

#define LOG_IDENT		"sew"
#define LOG_FACILITY	""
#define LOG_PRIORITY	LOG_AUTHPRIV

char *str_to_first_word(char *str)
{
	int i;
	char *mot;

	if (!str)
		return (NULL);

	mot = malloc ((1 + strlen (str)) * sizeof (char));
	memset (mot, 0, strlen (str) + 1);

	for (i = 0 ; *str != '\0' && !isspace(*str); i ++, str ++)
		mot[i] = *str;

	return (mot);
}

char **str_to_wordtab(char *str ,char sep)
{
	char **n;
	int mot, lettre, i;
	mot = lettre = i = 0;

	n = malloc (strlen(str) * sizeof(char *));

	while (*(str + i)) {
		if (*(str + i) == sep) {
			if (lettre) {
				lettre = n[mot][lettre] = 0;
				mot ++;
			}
		} else {
			if (lettre == 0) {
				n[mot] = realloc(NULL, 2);
				n[mot][0] = *(str + i);
				n[mot][1] = 0;
				lettre ++;
			} else {
				n[mot] =  realloc(n[mot], lettre + 2);
				n[mot][lettre] = *(str + i);
				n[mot][lettre + 1] = 0;
				lettre ++;
			}
		}
		i ++;
	}
	n[mot + 1] = NULL;
	return (n);
}

char *wordtab_to_str(char **tab)
{
	char *s, *n = NULL;
	char **t;
	int tl;

	if (*tab == NULL) {
		return (NULL);	
	}

	n = strdup (*tab);
	tl = strlen (n) + 2;

	for (t = (tab + 1); *t; t++) {
		s = n;
		n[tl - 2] = ' ';
		n[tl - 1] = '\0';

		tl += strlen(*t) + 1;
		if ((n = realloc (s, tl)) != NULL) {
			strcat (n, *t);
		} else {
			n = s;
		}
	}

	return (n);
}

void free_tab (char **ptr)
{
	int i = 0;
	while (*(ptr + i)) {
		free(*(ptr + i));
		i ++;
	}
	free(ptr);
	return ;
}

int main(int argc, char **argv, char **envp)
{
	char *cmd = NULL;
	char buffer [1024];
	struct passwd *user;
	FILE *fconf;
	char **tabcmd = NULL, **tabline, **s;
	int len1, len2, exec_ok;

	openlog(LOG_IDENT, LOG_PID | LOG_NDELAY, LOG_PRIORITY);

	if (argc == 1) { 
		cmd = strdup(argv[0]);
	} else {
		tabcmd = str_to_wordtab(argv[2], ' ');
		cmd = *tabcmd;
	}

	syslog (LOG_INFO, "Execution of %s asked by uid: %i.\n", cmd, getuid());

	if ((user = getpwuid (getuid())) == NULL)
		goto bad;

	if ((fconf = fopen (SEV_FILE, "r")) == NULL)
		goto bad;

	if (cmd == NULL)
		goto bad;

	while ((fgets(buffer, sizeof(buffer) - 1, fconf)) != NULL) {
		buffer[strlen(buffer) - 1] = 0;
		tabline = str_to_wordtab (buffer, ':');

		if (*tabline && *(tabline + 1)) {
			if (strcmp (*tabline, user->pw_name) == 0) {
				for (s = tabline + 2 ; *s ; s ++) {
					len1 = strlen (cmd);	
					len2 = strlen (*s);
					exec_ok = 0;

					if (strcmp (*s, ALL) == 0) {
						exec_ok = 1;
					}

					if (len1 >= len2 
					    && strcmp (*s, cmd + len1 - len2) == 0) {
						exec_ok = 1;
					}

					if (exec_ok == 1) {
						argv[0] = strdup (*(tabline + 1));
						cmd = wordtab_to_str (argv);

						syslog (LOG_INFO, "Execution for %s of [%s] allowed.\n",
							user->pw_name, cmd);

						free_tab (tabline);
						if (tabcmd != NULL)
							free_tab (tabcmd);

						closelog ();

						return execve (*argv, argv, envp);
					}
				}
			}
		}
		free_tab (tabline);
	}


bad:
	syslog(LOG_INFO, "No %s found, or access denied.\n", SEV_FILE);
	closelog();
	printf(DEF_DENY_MSG);
	return (-1);
}
